#!/bin/sh
#  _    _          ___           __ _     (R)
# | |  (_)_ _____ / __|___ _ _  / _(_)__ _
# | |__| \ V / -_) (__/ _ \ ' \|  _| / _` |
# |____|_|\_/\___|\___\___/_||_|_| |_\__, |
#                                    |___/
# Copyright (c) 2009-2021 LiveConfig GmbH.
# ----------------------------------------------------------------------------
# common/tools/lcservice.sh
# 
# LiveConfig service script, mainly used for system configuration upgrades
# Optional environment variable:
#   HTDOCS - alternative name for "htdocs" subdirectory (eg. "html")
#   RECURSIVE - set to "1" to fix permissions recursively
# ----------------------------------------------------------------------------

set -e

# . /etc/liveconfig/sysconfig || exit 0
LC_WEBROOT=/var/www

COMMAND=$1
RETVAL=0
HTDOCS=${HTDOCS:-htdocs}

# ----------------------------------------------------------------------------
# return user name of web server user
# ----------------------------------------------------------------------------
get_web_user() {
    WEB_USER=www-data
    [ -f /etc/SuSE-release ] && WEB_USER=wwwrun
    [ -f /etc/redhat-release ] && WEB_USER=apache
    [ -f /etc/gentoo-release ] && WEB_USER=apache
    [ "`uname -s`" = "FreeBSD" ] && WEB_USER=www
    return 0
}

# ----------------------------------------------------------------------------
# return group name of web server user
# ----------------------------------------------------------------------------
get_web_group() {
    WEB_GROUP=www-data
    [ -f /etc/SuSE-release ] && WEB_GROUP=www
    [ -f /etc/redhat-release ] && WEB_GROUP=apache
    [ -f /etc/gentoo-release ] && WEB_GROUP=apache
    [ "`uname -s`" = "FreeBSD" ] && WEB_GROUP=www
    return 0
}

# ----------------------------------------------------------------------------
# fix_awstats: fix configuration file /etc/awstats/liveconfig/awstats.conf.liveconfig
# ----------------------------------------------------------------------------
fix_awstats() {
    if [ -f /etc/awstats/liveconfig/awstats.conf.liveconfig ]; then
        sed -i -e 's/^LogFormat=4/LogFormat=1/g' /etc/awstats/liveconfig/awstats.conf.liveconfig
    fi
}

# ----------------------------------------------------------------------------
# fix_permissions: fix directory permissions in /var/www/*
# ----------------------------------------------------------------------------
fix_permissions() {
    SUBDIRS=${1:-*}
    get_web_user
    get_web_group
    echo -n "[LiveConfig] Fixing directory permissions in $LC_WEBROOT/$SUBDIRS"
    [ "x$RECURSIVE" = "x1" ] && echo -n " (recursion=ON)"
    echo -n ": "
    for i in $LC_WEBROOT/$SUBDIRS; do
        [ -d "$i/$HTDOCS" ] || continue
        # get lower-case directory name for user/group name
        USR="$(basename $i | tr '[:upper:]' '[:lower:]')"
        if ! getent passwd "$USR" >/dev/null; then
            echo "User '$USR' not found - not updating '$i'!" >&2
            continue
        fi
        echo -n "."
        chown root:root "$i"
        chmod 0755 "$i"

        # fix "apps" directory
        if [ -d "$i/apps" ]; then
            [ "x$RECURSIVE" = "x1" ] && chown -R "$USR":"$USR" "$i/apps"
            chown "$USR":"$WEB_GROUP" "$i/apps"
            chmod 0750 "$i/apps"
            chown -f "$USR":"$USR" "$i/apps/*" || true
            chmod -f 0755 "$i/apps/*" || true
        fi

        # fix "conf" directory
        if [ -d "$i/conf" ]; then
            chown "$WEB_USER":"$USR" "$i/conf"
            chmod 0750 "$i/conf"
            [ -f "$i/conf/php5/php-fcgi-starter" ] && chattr +i "$i/conf/php5/php-fcgi-starter"
            [ -f "$i/conf/php5/php.ini" ] && chattr +i "$i/conf/php5/php.ini"
        fi

        # fix "htdocs" directory
        if [ -d "$i/$HTDOCS" ]; then
            [ "x$RECURSIVE" = "x1" ] && chown -R "$USR":"$USR" "$i/$HTDOCS"
            chown "$USR":"$WEB_GROUP" "$i/$HTDOCS"
            chmod 0750 "$i/$HTDOCS"
            # move "cgi-bin" to "htdocs/cgi-bin"
            if [ -d "$i/cgi-bin" ] && [ ! -d "$i/$HTDOCS/cgi-bin" -o "$(ls -A '$i/$HTDOCS/cgi-bin')" ]; then
                mv "$i/cgi-bin" "$i/$HTDOCS/"
            fi
            # fix permissions of cgi-bin:
            if [ -d "$i/$HTDOCS/cgi-bin" ]; then
                chown "$USR":"$USR" "$i/$HTDOCS/cgi-bin"
                chmod 0755 "$i/$HTDOCS/cgi-bin"
            fi
        fi

        # fix "logs" directory
        if [ -d "$i/logs" ]; then
            chown "$WEB_USER":"$USR" "$i/logs"
            chmod 0750 "$i/logs"
            chmod -f 0644 "$i/logs/access.log" "$i/logs/error.log" || true
            if [ ! -d "$i/logs/priv" ]; then
                mkdir "$i/logs/priv"
                chown "$WEB_USER":"$USR" "$i/logs/priv"
                chmod 0770 "$i/logs/priv"
            fi
        fi

        # fix "priv" directory
        if [ -d "$i/priv" ]; then
            [ "x$RECURSIVE" = "x1" ] && chown -R "$USR":"$USR" "$i/priv"
            chown "$USR":"$USR" "$i/priv"
            chmod 0750 "$i/priv"
        fi

        # fix all other directories
        [ -d "$i/stats" ] && chmod 0750 "$i/stats"
        [ -d "$i/tmp" ] && chown "$USR":"$WEB_GROUP" "$i/tmp" && chmod 01770 "$i/tmp"
    done
    echo " - done."
    return 0
}

# ----------------------------------------------------------------------------
# check_postfix_domains: check if all domains are present in virtual_domains
# ----------------------------------------------------------------------------
check_postfix_domains() {
    [ -f /etc/postfix/virtual_alias -a -f /etc/postfix/virtual_domains ] || return 0
    echo -n "[LiveConfig] Checking domains in Postfix' virtual_domains file: "
    names=`grep '^[^#]' /etc/postfix/virtual_alias | cut -f 1 | cut -d '@' -f 2 | sort | uniq`;
    for name in $names; do
        grep -q "$name" /etc/postfix/virtual_domains || (echo "$name\t$name" >> /etc/postfix/virtual_domains; echo -n "$name ");
    done
    # modify (existing) virtual_domains to create a hash map from it:
    sed -i -e 's/^\([^#][^\t]*\)$/\1\t\1/' /etc/postfix/virtual_domains
    echo "- done."
    postmap /etc/postfix/virtual_domains
    return 0
}

# ----------------------------------------------------------------------------
# fix_options: fix "FollowSymLinks" option in .htaccess files and Apache config
# (replace with "SymLinksIfOwnerMatch")
# ----------------------------------------------------------------------------
fix_options() {
    echo "Fixing FollowSymLinks options:"
    # fix all .htaccess files
    find "$LC_WEBROOT" -type f -name .htaccess \
        -exec grep -il "^[[:space:]]*options.*followsymlinks" {} \; \
        -exec sed -i -e 's/^\(\s*options.*\)followsymlinks/\1SymLinksIfOwnerMatch/i' {} \;
    # fix vHost configs
    if test -d "/etc/apache2/sites-available"; then
      CFGDIR="/etc/apache2/sites-available"
    elif test -d "/etc/apache2/vhosts.d"; then
      CFGDIR="/etc/apache2/vhosts.d"
    elif test -d "/etc/httpd/vhosts.d"; then
      CFGDIR="/etc/httpd/vhosts.d"
    fi
    if test -z "$CFGDIR"; then
        echo "WARNING: Apache vHost configuration directory not found!"
    else
        find "$CFGDIR" -type f -name "*.conf" \
            -exec grep -iq "created by LiveConfig" {} \; \
            -exec grep -il "FollowSymLinks,SymLinksIfOwnerMatch" {} \; \
            -exec sed -i -e 's/FollowSymLinks,SymLinksIfOwnerMatch/SymLinksIfOwnerMatch/' {} \;
    fi
    echo "- done."
    return 0
}

# ----------------------------------------------------------------------------
# fix_symlinks: fix symbolic links in Apache vHost configuration
# ----------------------------------------------------------------------------
fix_symlinks() {
    if [ -d /etc/apache2/sites-available -a -d /etc/apache2/sites-enabled ]; then
        cd /etc/apache2/sites-enabled
        for i in *.conf; do
            if [ -f "../sites-available/$i" -a ! -h "$i" ]; then
                echo "fixing symlink for $i"
                mv $i ../sites-available && ln -s ../sites-available/$i $i
            fi
        done
    fi
}

# ----------------------------------------------------------------------------
case "$COMMAND" in
    fix-awstats)
        fix_awstats
        ;;
    fix-permissions)
        fix_permissions $2
        ;;
    check-postfix-domains)
        check_postfix_domains
        ;;
    fix-options)
        fix_options
        ;;
    fix-symlinks)
        fix_symlinks
        ;;
    *)
        echo "Usage: lcservice.sh {fix-permissions|check-postfix-domains|fix-options|fix-symlinks}"
        RETVAL=1
        ;;
esac
exit $RETVAL

# <EOF>-----------------------------------------------------------------------
