--  _    _          ___           __ _     (R)
-- | |  (_)_ _____ / __|___ _ _  / _(_)__ _
-- | |__| \ V / -_) (__/ _ \ ' \|  _| / _` |
-- |____|_|\_/\___|\___\___/_||_|_| |_\__, |
--                                    |___/
-- Copyright (c) 2009-2019 LiveConfig GmbH.
-- ---------------------------------------------------------------------------
-- common/lua/spamassassin.lua
-- SpamAssassin configuration/management
-- $Id$
-- ---------------------------------------------------------------------------

-- Imported functions
local LC = LC
local LCS = LCS
local os = os
local io = io
local ipairs = ipairs
local string = string

-- Module declaration
module("spamassassin")

-- Exported functions
--   detect()
--   install()
--   uninstall()
--   editMailbox()
--   deleteMailbox()

-- ---------------------------------------------------------------------------
-- detect()
--
-- Detect installed SpamAssassin package
-- ---------------------------------------------------------------------------
function detect()
  local data
  local mod

  -- check if smtp server software was already detected; avoid double-detection
  -- also across multiple threads
  LC.mutex.lock("spamassassin.detect")

  if LCS.spamassassin ~= nil and LCS.spamassassin.data ~= nil then
    -- spamassassin software already detected; load values from global LCS storage
    LC.mutex.unlock("spamassassin.detect")
    return LCS.spamassassin.data
  end

  -- package is known as "spamassassin" on all distributions
  local pkg, v = LC.distribution.hasPackage('spamassassin')
  if pkg ~= nil then
    LC.log.print(LC.log.DEBUG, "Found package '", pkg, "' (Version ", v, ")")
    data = {
      pkgname       = pkg;
      pkgversion    = v;
      configured    = false;
      configpath    = "/etc/spamassassin";
      sa_start_cmd  = "/etc/init.d/spamassassin start";
      sa_stop_cmd   = "/etc/init.d/spamassassin stop";
      optionsfile   = "/etc/default/spamassassin";
    }
    if LC.distribution.family == "Debian" then
      if LC.fs.is_file("/etc/default/spamd") then
        -- new name since Debian 12
        data.sa_start_cmd = "/usr/bin/systemctl start spamd"
        data.sa_stop_cmd = "/usr/bin/systemctl stop spamd"
        data.optionsfile = "/etc/default/spamd"
      end
    elseif LC.distribution.family == "RedHat" then
      data.configpath   = "/etc/mail/spamassassin"
      data.optionsfile  = "/etc/sysconfig/spamassassin";
      data.sa_start_cmd = "/sbin/service spamassassin start"
      data.sa_stop_cmd  = "/sbin/service spamassassin stop"
    elseif LC.distribution.family == "SUSE" then
      data.configpath   = "/etc/mail/spamassassin"
      data.optionsfile  = "/etc/sysconfig/spamd"
      data.sa_start_cmd = "/sbin/service spamd start"
      data.sa_stop_cmd  = "/sbin/service spamd stop"
    end

    if LC.distribution.family == "RedHat" or LC.distribution.family == "SUSE" then
      data.configpath   = "/etc/mail/spamassassin"
      if LC.distribution.family == "SUSE" then
        data.optionsfile    = "/etc/sysconfig/spamd";
      else
        data.optionsfile  = "/etc/sysconfig/spamassassin";
      end
    end
    -- check for status file
    local statusfile = data.configpath .. "/liveconfig.status"
    if LC.fs.is_file(statusfile) then
      local r = LC.liveconfig.readStatus(statusfile)
      if r and r.revision then
        data.configured = true
      end
    end
    LCS.spamassassin = { }
    LCS.spamassassin.data = data
  end

  LC.mutex.unlock("spamassassin.detect")
  
  return data
end

-- ---------------------------------------------------------------------------
-- install()
--
-- Install (configure) SpamAssassin and LiveConfig SpamAssassin Milter (lcsam)
-- ---------------------------------------------------------------------------
function install()

  -- get configuration
  local cfg = detect()
  if cfg == nil then
    return false, "SpamAssassin not found"
  end

  if cfg.configured then
    -- already configured...
    return true
  end

  -- acquire mutex
  LC.mutex.lock("spamassassin.install")

  -- stop SpamAssassin (if running)
  -- (must be done *NOW*, because location of some files might change)
  LC.exec(cfg.sa_stop_cmd)

  -- create system group "spamd" if it doesn't exist
  if not LC.sys.group_exists("spamd") then
    LC.log.print(LC.log.INFO, "Creating system group 'spamd'")
    LC.exec("/usr/sbin/groupadd -r spamd")
  end

  -- create system user "spamd" if it doesn't exist
  if not LC.sys.user_exists("spamd") then
    LC.log.print(LC.log.INFO, "Creating system user 'spamd'")
    LC.exec("useradd -r -d /var/lib/spamassassin -g spamd -s /bin/false spamd")
  end

  -- create per-user configuration directory for SpamAssassin
  if not LC.fs.is_dir("/var/lib/spamassassin") then
    LC.log.print(LC.log.INFO, "Creating directory '/var/lib/spamassassin'")
    LC.fs.mkdir("/var/lib/spamassassin")
    -- adjust owner
    LC.fs.setperm("/var/lib/spamassassin", 750, "spamd", "root")
  end

  local spamass_file = "/etc/postfix/spamassassin"
  if not LC.fs.is_file(spamass_file .. ".db") then
    if not LC.fs.is_file(spamass_file) then
      local fha, msg = io.open(spamass_file, "w")
      if fha == nil then
        LC.log.print(LC.log.ERR, "Can't create '" .. spamass_file .. "' for writing: ", msg)
      else
        LC.liveconfig.writeHeader(fha)
      end
      fha:close()
      -- adjust permissions
      LC.fs.setperm(spamass_file, 640, "root", "spamd")
    end

    -- update map file
    LC.exec("/usr/sbin/postmap " .. spamass_file)
  end

  -- create backup of startup options file
  if LC.fs.is_file(cfg.optionsfile) and not LC.fs.is_file(cfg.optionsfile .. ".lcbak") then
    LC.fs.copy(cfg.optionsfile, cfg.optionsfile .. ".lcbak")
  end

  -- patch SpamAssassin start options
  if LC.distribution.family == "Debian" then
    LC.exec([[sed -i \
-e 's/^ENABLED=0/ENABLED=1/' \
-e 's/^OPTIONS="\(--create-prefs.*\)"/#OPTIONS="\1"\nOPTIONS="-m 5 -H --socketpath=\/var\/run\/spamd.sock --socketowner=root --socketgroup=spamd --socketmode=0660 -x --virtual-config-dir=\/var\/lib\/spamassassin\/%u\/ -u spamd"/' \
]] .. cfg.optionsfile)
  elseif LC.distribution.family == "RedHat" then
    LC.exec([[sed -i \
-e 's/^SPAMDOPTIONS="\(\(-d \)\?-c.*\)"/#SPAMDOPTIONS="\1"\nSPAMDOPTIONS="\2-m 5 -H --socketpath=\/var\/run\/spamd.sock --socketowner=root --socketgroup=spamd --socketmode=0660 -x --virtual-config-dir=\/var\/lib\/spamassassin\/%u\/ -u spamd"/' \
]] .. cfg.optionsfile)
  elseif LC.distribution.family == "SUSE" then
    LC.exec([[sed -i \
-e 's/^SPAMD_ARGS="\(-d -c.*\)"/#SPAMD_ARGS="\1"\nSPAMD_ARGS="-d -m 5 -H --socketpath=\/var\/run\/spamd.sock --socketowner=root --socketgroup=spamd --socketmode=0660 -x --virtual-config-dir=\/var\/lib\/spamassassin\/%u\/ -u spamd"/' \
]] .. cfg.optionsfile)
  end

  -- start SpamAssassin
  LC.log.print(LC.log.INFO, "Starting SpamAssassin...")
  LC.exec(cfg.sa_start_cmd)

  -- install service (systemd)
  if LC.fs.is_dir("/lib/systemd/system") and not LC.fs.is_file("/lib/systemd/system/lcsam.service") then
    LC.exec("cp --preserve=all " .. LC.liveconfig.libexecdir .. "/lcsam.service /lib/systemd/system/")
    if LC.fs.is_file("/bin/systemctl") then
      LC.exec("/bin/systemctl enable lcsam")
    end
  end

  -- lcsam: install init script
  if not LC.fs.is_file("/etc/init.d/lcsam") then
    LC.exec("cp " .. LC.liveconfig.libexecdir .. "/lcsam.init /etc/init.d/lcsam")
    LC.fs.setperm("/etc/init.d/lcsam", "0700", "root", "root")

    if LC.fs.is_file("/usr/sbin/update-rc.d") then
      LC.exec("/usr/sbin/update-rc.d lcsam defaults")
    elseif LC.fs.is_file("/sbin/chkconfig") then
      LC.exec("/sbin/chkconfig --add lcsam")
    elseif LC.fs.is_file("/sbin/rc-update") then
      LC.exec("/sbin/rc-update add lcsam default")
    else
      LC.log.print(LC.log.ERR, "Can't install init script 'lcsam' - no supported tool found")
    end
  end

  -- start lcsam
  if LC.fs.is_file("/sbin/service") then
    LC.exec("/sbin/service lcsam start")
  else
    LC.exec("/etc/init.d/lcsam start")
  end

  -- write status file
  LC.liveconfig.writeStatus(cfg.configpath .. "/liveconfig.status", "spamassassin", 0, os.date("%Y-%m-%d %H:%M:%S %Z"))

  -- set "configured" flag
  cfg.configured = true

  -- release mutex
  LC.mutex.unlock("spamassassin.install")

  return true

end

-- ---------------------------------------------------------------------------
-- uninstall()
--
-- Uninstall/disable LiveConfig SpamAssassin Milter (lcsam)
-- ---------------------------------------------------------------------------
function uninstall()

  -- get configuration
  local cfg = detect()
  if cfg == nil then
    return false, "SpamAssassin not found"
  end

  -- acquire mutex
  LC.mutex.lock("spamassassin.install")

  -- quit lcsam
  if LC.fs.is_file("/sbin/service") then
    LC.exec("/sbin/service lcsam stop")
  else
    LC.exec("/etc/init.d/lcsam stop")
  end

  -- disable lcsam
  if LC.fs.is_file("/lib/systemd/system/lcsam") then
    -- disable service (systemd)
    if LC.fs.is_file("/bin/systemctl") then
      LC.exec("/bin/systemctl disable lcsam")
    end
    -- remove service file
    os.remove("/lib/systemd/system/lcsam")
  end

  if LC.fs.is_file("/etc/init.d/lcsam") then
    if LC.fs.is_file("/usr/sbin/update-rc.d") then
      LC.exec("/usr/sbin/update-rc.d -f lcsam remove")
    elseif LC.fs.is_file("/sbin/chkconfig") then
      LC.exec("/sbin/chkconfig --del lcsam")
    elseif LC.fs.is_file("/sbin/rc-update") then
      LC.exec("/sbin/rc-update del lcsam default")
    else
      LC.log.print(LC.log.ERR, "Can't uninstall init script 'lcsam' - no supported tool found")
    end
    os.remove("/etc/init.d/lcsam")
  end

  -- remove status file
  local statusfile = cfg.configpath .. "/liveconfig.status"
  if LC.fs.is_file(statusfile) then
    os.remove(statusfile)

    -- remove spamassassin user databases (only if lcsam was configured!)
    if LC.fs.is_file("/etc/postfix/spamassassin") then
      os.remove("/etc/postfix/spamassassin")
    end
    if LC.fs.is_file("/etc/postfix/spamassassin.db") then
      os.remove("/etc/postfix/spamassassin.db")
    end

  end

  -- quit SpamAssassin
  LC.log.print(LC.log.INFO, "Stopping SpamAssassin...")
  LC.exec(cfg.sa_stop_cmd)

  -- restore SpamAssassin startup configuration
  if LC.fs.is_file(cfg.optionsfile .. ".lcbak") then
    LC.fs.rename(cfg.optionsfile .. ".lcbak", cfg.optionsfile)
  end

  -- release mutex
  LC.mutex.unlock("spamassassin.install")

end

-- ---------------------------------------------------------------------------
-- editMailbox()
--
-- Add/edit e-mail address to/in SpamAssassin database
-- ---------------------------------------------------------------------------
function editMailbox(search, addrs, usercfg)

  -- get configuration
  local cfg = detect()
  if cfg == nil or not cfg.configured then
    -- SpamAssassin not installed or not configured
    return false
  end

  -- acquire mutex
  LC.mutex.lock("spamassassin.edit")

  -- update SpamAssassin database file
  local spamass_file = "/etc/postfix/spamassassin"
  local fhr, fhw, msg, i, s
  fhr, msg = io.open(spamass_file, "r")
  if fhr == nil then
    LC.log.print(LC.log.ERR, "Can't open '", spamass_file, "' for reading: ", msg)
    LC.mutex.unlock("spamassassin.edit")
    return false, "Can't open '" .. spamass_file .. "' for reading: " .. msg
  end

  fhw, msg = io.open(spamass_file .. ".tmp", "w")
  if fhw == nil then
    LC.log.print(LC.log.ERR, "Can't open '", spamass_file .. ".tmp", "' for writing: ", msg)
    fhr:close()
    LC.mutex.unlock("spamassassin.edit")
    return false, "Can't open '" .. spamass_file .. ".tmp" .. "' for writing: " .. msg
  end
  -- adjust owner & permissions - only Postfix needs to read this file:
  LC.fs.setperm(spamass_file .. ".tmp", 640, "root", "spamd")

  while true do
    line = fhr:read()
    if line == nil then break end
    found = false
    for i, s in ipairs(search) do
      if string.find(line, s) ~= nil then
        found = true
        break
      end
    end
    if found == false then
      -- only write line if not matched
      fhw:write(line, "\n")
    end
  end
  fhr:close()

  -- if spam analysis is enabled, add all addresses/aliases:
  if usercfg then
    for i, s in ipairs(addrs) do
      if usercfg.subjprefix == nil then usercfg.subjprefix = "" end
      fhw:write(s, "\t1 ", usercfg.warn, " ", usercfg.reject, " 1 ", usercfg.userdir, " ", usercfg.subjprefix, "\n")
    end
  end

  fhw:close()
  LC.fs.rename(spamass_file .. ".tmp", spamass_file)

  -- update map file
  LC.timeout.set('LC.spamassassin.update_map', 10, 60)

  -- update blacklist/whitelist settings
  local userprefs_path = "/var/lib/spamassassin/" .. usercfg.userdir
  local userprefs_file = userprefs_path .. "/user_prefs"
  if not LC.fs.is_dir(userprefs_path) then
    LC.fs.mkdir(userprefs_path)
    LC.fs.setperm(userprefs_path, 750, "spamd", "root")
  end
  if (usercfg.blacklist == nil and usercfg.whitelist == nil) then
    -- just remove user_prefs file
    os.remove(userprefs_file)
  else
    -- write/update user_prefs:
    fhw, msg = io.open(userprefs_file .. ".tmp", "w")
    if fhw == nil then
      LC.log.print(LC.log.ERR, "Can't open '", userprefs_file .. ".tmp", "' for writing: ", msg)
      fhr:close()
      LC.mutex.unlock("spamassassin.edit")
      return false, "Can't open '" .. userprefs_file .. ".tmp" .. "' for writing: " .. msg
    end
    -- adjust owner & permissions - only Postfix needs to read this file:
    LC.fs.setperm(userprefs_file .. ".tmp", 640, "spamd", "root")

    -- write whitelist settings
    if usercfg.whitelist ~= nil then
      for i, s in ipairs(usercfg.whitelist) do
        fhw:write("whitelist_from ", s, "\n")
      end
    end

    -- write blacklist settings
    if usercfg.blacklist ~= nil then
      for i, s in ipairs(usercfg.blacklist) do
        fhw:write("blacklist_from ", s, "\n")
      end
    end

    fhw:close()
    LC.fs.rename(userprefs_file .. ".tmp", userprefs_file)

  end

  -- release mutex
  LC.mutex.unlock("spamassassin.edit")

end

-- ---------------------------------------------------------------------------
-- deleteMailbox()
--
-- Delete e-mail address from SpamAssassin database
-- ---------------------------------------------------------------------------
function deleteMailbox(search, userdir)

  -- get configuration
  local cfg = detect()
  if cfg == nil then
    return false, "SpamAssassin not found"
  end

  -- acquire mutex
  LC.mutex.lock("spamassassin.edit")

  -- update SpamAssassin database file
  local spamass_file = "/etc/postfix/spamassassin"
  local fhr, fhw, msg, i, s
  fhr, msg = io.open(spamass_file, "r")
  if fhr == nil then
    LC.log.print(LC.log.ERR, "Can't open '", spamass_file, "' for reading: ", msg)
    LC.mutex.unlock("spamassassin.edit")
    return false, "Can't open '" .. spamass_file .. "' for reading: " .. msg
  end

  fhw, msg = io.open(spamass_file .. ".tmp", "w")
  if fhw == nil then
    LC.log.print(LC.log.ERR, "Can't open '", spamass_file .. ".tmp", "' for writing: ", msg)
    fhr:close()
    LC.mutex.unlock("spamassassin.edit")
    return false, "Can't open '" .. spamass_file .. ".tmp" .. "' for writing: " .. msg
  end
  -- adjust owner & permissions - only Postfix needs to read this file:
  LC.fs.setperm(spamass_file .. ".tmp", 640, "root", "spamd")

  while true do
    line = fhr:read()
    if line == nil then break end
    found = false
    for i, s in ipairs(search) do
      if string.find(line, s) ~= nil then
        found = true
        break
      end
    end
    if found == false then
      -- only write line if not matched
      fhw:write(line, "\n")
    end
  end
  fhr:close()

  fhw:close()
  LC.fs.rename(spamass_file .. ".tmp", spamass_file)

  -- update map file
  LC.timeout.set('LC.spamassassin.update_map', 10, 60)

  -- delete SpamAssassin user_prefs (blacklist/whitelist)
  local userprefs_path = "/var/lib/spamassassin/" .. userdir
  if (LC.fs.is_dir(userprefs_path)) then
    os.execute("rm -f " .. userprefs_path .. "/*")
    LC.fs.rmdir(userprefs_path)
  end

  -- release mutex
  LC.mutex.unlock("spamassassin.edit")

end

-- ---------------------------------------------------------------------------
-- update_map()
--
-- Update "spamassassin.db" map file
-- ---------------------------------------------------------------------------
function update_map()
  LC.log.print(LC.log.DEBUG, "LC.spamassassin.update_map() called")

  -- update map file
  LC.exec("/usr/sbin/postmap /etc/postfix/spamassassin")

end

-- <EOF>----------------------------------------------------------------------
